#########################################################################
#
#	Ensemble approach for Intrusion Detection System (IDS)  
#	Created: 04/05/2012
#	Author: Alexandre Balon-Perin
#
#########################################################################

source("CV.R")
	  	  	
#########################################################################
#
#						INITIALIZATION
#
#########################################################################
#rm(list=ls())
#setwd('/Users/alex/University/Master Thesis/Code/')
#par(mfrow=c(1,2),ask=TRUE)
#par(ask=TRUE)
op <- options(digits=10)
options(op)

library(lazy)
library(randomForest)
library(tree)
library(e1071)
library(rpart)


#	Read the training data
######################################################################
#!!! Uncomment these lines if the following variables are not in memory yet
#!!! It takes a few minutes to read these files (dos and normal mainly) so 
#!!! it is better to run this part separately
#!!! To produce the files kddcup_*.data, run the script "create_dataset_IDS.R" 
#!!! after downloading the training set from http://kdd.ics.uci.edu/databases/kddcup99/kddcup99.html

#r2l <- read.table("kddcup_r2l.data",sep=",",header=T)
#u2r <- read.table("kddcup_u2r.data",sep=",",header=T)
#probe_original <- read.table("kddcup_probe.data",sep=",",header=T)
#dos <- read.table("kddcup_dos.data",sep=",",header=T)
#normal <- read.table("kddcup_normal.data",sep=",",header=T)

#probe <- probe_original[sample(1:dim(probe_original)[1],10000),]

#	Features selected by the three algorithms svm, lgp and mars
#	for the different classes of attack
########################################################################

features.r2l.svm <- c("srv_count","service","duration","count","dst_host_count")
features.r2l.lgp <- c("is_guest_login","num_access_files","dst_bytes","num_failed_logins","logged_in")
features.r2l.mars <- c("srv_count","service","dst_host_srv_count","count","logged_in")
features.r2l.combined <- unique(c(features.r2l.svm,features.r2l.lgp,features.r2l.mars))

features.probe.svm <- c("src_bytes","dst_host_srv_count","count","protocol_type","srv_count")
features.probe.lgp <- c("srv_diff_host_rate","rerror_rate","dst_host_diff_srv_rate","logged_in","service")
features.probe.mars <- c("src_bytes","dst_host_srv_count","dst_host_diff_srv_rate","dst_host_same_srv_rate","srv_count")
features.probe.combined <- unique(c(features.probe.svm,features.probe.lgp,features.probe.mars))

features.u2r.svm <- c("src_bytes","duration","protocol_type","logged_in","flag")
features.u2r.lgp <- c("root_shell", "dst_host_srv_serror_rate","num_file_creations","serror_rate","dst_host_same_src_port_rate")
features.u2r.mars <- c("dst_host_srv_count","duration","count","srv_count","dst_host_count")
features.u2r.combined <- unique(c(features.u2r.svm,features.u2r.lgp,features.u2r.mars))

features.normal.svm <- c("dst_bytes","dst_host_count","logged_in","dst_host_same_srv_rate","flag")
features.normal.lgp <- c("dst_bytes","src_bytes","dst_host_rerror_rate","num_compromised","hot")
features.normal.mars <- c("dst_bytes","src_bytes","logged_in","service","hot")
features.normal.combined <- unique(c(features.normal.svm,features.normal.lgp,features.normal.mars))



print("data acquired...")

#########################################################################
#
#						MODEL SELECTION	
#
#########################################################################
model<-function(){
	
	print('Model Selection')
	
	#######################################################################
	#
	#							PREPARATIONS
	#
	#######################################################################
	print('Preparation...')
	time1 <-Sys.time()
	
	
	#		Sample the training set containing the normal labels
	#		for each class of attack in order to have the same number 
	#		of training data belonging to the "normal" class and the 
	#		"attack" class
	#######################################################################
	
	normal_sample.r2l <- normal[sample(1:dim(normal)[1],dim(r2l)[1]),]
	normal_sample.probe <- normal[sample(1:dim(normal)[1],dim(probe)[1]),]
	normal_sample.u2r <- normal[sample(1:dim(normal)[1],dim(u2r)[1]),]
	#normal_sample.dos <- normal[sample(1:dim(normal)[1],dim(dos)[1]),]
	
	
	# This part was useful before the separation normal/attack because 
	# attack was composed of different types for each class
	######################################################################
	
	normal.r2l.Y <- matrix(normal_sample.r2l[,c("label")]) 
	normal.probe.Y <- matrix(normal_sample.probe[,c("label")]) 
	normal.u2r.Y <- matrix(normal_sample.u2r[,c("label")]) 
	#normal.dos.Y <- matrix(normal_sample.dos[,c("label")]) 
	
	#######################################################################
	#		Class of attack Remote to Local (r2l)
	#######################################################################
	
	#	Select the features according to the algorithms(svm,lgp and mars)
	#	for this particular type of attack. Combined contains the 
	#	combination of the features selected by the 3 algorithms  
	#######################################################################
	#features.r2l.svm <- c(features.r2l.svm,"label")
	r2l_svm <- r2l[,features.r2l.svm]
	r2l_lgp <- r2l[,features.r2l.lgp]
	r2l_mars <- r2l[,features.r2l.mars]
	r2l_combined <- r2l[,features.r2l.combined]
	r2l_ALL <- r2l[,colnames(r2l) != "label"]
	
	r2l.Y <- matrix(r2l[,c("label")])
	r2l.Y[,1] = "attack"

	

	#	Merge the "normal" instances and the "r2l" instances and shuffle the result
	############################################################################### 

	#r2l_svm[,"label"] = "attack"
	
	r2l_svm.tr <- rbind(normal_sample.r2l[,features.r2l.svm],r2l_svm)
	r2l_svm.tr <- r2l_svm.tr[sample(1:nrow(r2l_svm.tr),replace=F),]
	r2l_lgp.tr <- rbind(normal_sample.r2l[,features.r2l.lgp],r2l_lgp)
	r2l_lgp.tr <- r2l_lgp.tr[sample(1:nrow(r2l_lgp.tr),replace=F),]
	r2l_mars.tr <- rbind(normal_sample.r2l[,features.r2l.mars],r2l_mars)
	r2l_mars.tr <- r2l_mars.tr[sample(1:nrow(r2l_mars.tr),replace=F),]
	r2l_ALL.tr <- rbind(normal_sample.r2l[,colnames(normal_sample.r2l) != "label"],r2l_ALL)
	r2l_ALL.tr <- r2l_ALL.tr[sample(1:nrow(r2l_ALL.tr),replace=F),]

	r2l.Y.tr <- rbind(normal.r2l.Y,r2l.Y)
	r2l.Y.tr <- matrix(r2l.Y.tr[sample(1:nrow(r2l.Y.tr),replace=F),])
	
	
	#pos = which(r2l_svm.tr[,"label"] == "attack")
	#neg = which(r2l_svm.tr[,"label"] == "normal.")
	
	
	#plot(r2l_svm.tr[pos,features.r2l.svm[1]],r2l_svm.tr[pos,features.r2l.svm[2]],col="red" )
	#par(new="T")
	#plot(r2l_svm.tr[neg,features.r2l.svm[1]],r2l_svm.tr[neg,features.r2l.svm[2]],col="green" )
	
	#######################################################################
	#		Class of attack Probe
	#######################################################################
	
	#	Select the features according to the algorithms(svm,lgp and mars)
	#	for this particular type of attack. Combined contains the 
	#	combination of the features selected by the 3 algorithms  
	#######################################################################
	
	probe_svm <- probe[,features.probe.svm]
	probe_lgp <- probe[,features.probe.lgp]
	probe_mars <- probe[,features.probe.mars]
	probe_ALL <- probe[,colnames(probe) != "label"]
	
	probe.Y <- matrix(probe[,c("label")])
	#probe.Y[,1] = "attack"	
	
	#	Merge the "normal" instances and the "probe" instances and shuffle the result
	############################################################################### 
	
	probe_svm.tr <- rbind(normal_sample.probe[,features.probe.svm],probe_svm)
	probe_svm.tr <- probe_svm.tr[sample(1:nrow(probe_svm.tr),replace=F),]
	probe_lgp.tr <- rbind(normal_sample.probe[,features.probe.lgp],probe_lgp)
	probe_lgp.tr <- probe_lgp.tr[sample(1:nrow(probe_lgp.tr),replace=F),]
	probe_mars.tr <- rbind(normal_sample.probe[,features.probe.mars],probe_mars)
	probe_mars.tr <- probe_mars.tr[sample(1:nrow(probe_mars.tr),replace=F),]
	probe_ALL.tr <- rbind(normal_sample.probe[,colnames(normal_sample.probe) != "label"],probe_ALL)
	probe_ALL.tr <- probe_ALL.tr[sample(1:nrow(probe_ALL.tr),replace=F),]
	
	probe.Y.tr <- rbind(normal.probe.Y,probe.Y)
	probe.Y.tr <- matrix(probe.Y.tr[sample(1:nrow(probe.Y.tr),replace=F),])



	#######################################################################
	#		Class of attack User to Root (u2r)
	#######################################################################

	u2r_svm <- u2r[,features.u2r.svm]
	u2r_lgp <- u2r[,features.u2r.lgp]
	u2r_mars <- u2r[,features.u2r.mars]
	u2r.Y <- matrix(u2r[,c("label")])
	u2r.Y[,1] = "attack"

	#u2r_svm.tr <- rbind(normal_svm,u2r_svm)
	#u2r_lgp.tr <- rbind(normal_lgp,u2r_lgp)
	#u2r_mars.tr <- rbind(normal_mars,u2r_mars)
	#u2r.Y.tr <- rbind(normal.Y,u2r.Y)

	#######################################################################
	#		Class Normal 
	#		(correspond to normal traffic without attack)
	#######################################################################

	
	normal_svm <- normal_sample.r2l[,features.normal.svm]
	normal_lgp <- normal_sample.r2l[,features.normal.lgp]
	normal_mars <- normal_sample.r2l[,features.normal.mars]
	
	#normal_svm.tr <- rbind(normal_svm,r2l_svm,u2r_svm,probe_svm)
	#normal_lgp.tr <- rbind(normal_lgp,r2l_lgp,u2r_lgp,probe_lgp)
	#normal_mars.tr <- rbind(normal_mars,r2l_mars,u2r_mars,probe_mars)

	

	time2 <-Sys.time()
	computation.time<-time2-time1

	print("Preparation time:")
	print(time2-time1)
	
	#######################################################################
	#
	#		10-fold CROSS-VALIDATION to assess the models accuracy
	#
	#######################################################################	
	print("Cross-Validation...")
	time3 <- Sys.time()
	
	# CV for Remote to Local
	########################	
	#cv(r2l_svm.tr, r2l_lgp.tr, r2l_mars.tr, r2l_ALL.tr, r2l.Y.tr)
	
	# CV for Probe
	########################
	cv(probe_svm.tr, probe_lgp.tr, probe_mars.tr, probe_ALL.tr, probe.Y.tr)

	time4 <-Sys.time()

	print("Time to compute cross-validation:")
	print(time4-time3) 
	
	

}



#########################################################################
#
#								MAIN 	
# 
##########################################################################
#op <- options(digits.secs=2)
#time1 <-Sys.time()
#options(op)

model()
print('===========================================================================================================')

#time2 <-Sys.time()
#computation.time<-time2-time1

#print(paste("The computation time is : ", round(computation.time))) 

